/******************************************************************************
*  Nano-RK, a real-time operating system for sensor networks.
*  Copyright (C) 2007, Real-Time and Multimedia Lab, Carnegie Mellon University
*  All rights reserved.
*
*  This is the Open Source Version of Nano-RK included as part of a Dual
*  Licensing Model. If you are unsure which license to use please refer to:
*  http://www.nanork.org/nano-RK/wiki/Licensing
*
*  This program is free software: you can redistribute it and/or modify
*  it under the terms of the GNU General Public License as published by
*  the Free Software Foundation, version 2.0 of the License.
*
*  This program is distributed in the hope that it will be useful,
*  but WITHOUT ANY WARRANTY; without even the implied warranty of
*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*  GNU General Public License for more details.
*
*  You should have received a copy of the GNU General Public License
*  along with this program.  If not, see <http://www.gnu.org/licenses/>.
*
*  Contributing Authors (specific to this file):
*  Anthony Rowe
*  Zane Starr
*******************************************************************************/


#include "tmote_basic_sensor.h"
#include <nrk_driver_list.h>
#include <nrk_driver.h>
#include <include.h>
#include <stdio.h>
#include <nrk_error.h>
#include <nrk.h>
#include <stdint.h>
#include <nrk_timer.h>


#define CALC_LIGHT(value) (((1.5 * .769 * value)/ 4096) * 1000)


uint8_t BUTTON_STATUS ; 

static uint8_t mode;

uint8_t channel;
uint8_t is_open;

uint8_t set_mode_TEMP_sensor (uint8_t newMode);
   
/** @brief transmit a byte MSB first to the SHT11 */
static inline void sht11_send_byte(uint8_t byte)
{
   int i;
   for(i = 7; i >= 0; --i)
   {
      if (byte & (1 << i))
	 DATA_HI();
      else
	 DATA_LO();
      
      SCLK_HI();
      
      SCLK_LO();
   }
}

/** @brief receives one byte of data from the SHT11*/
static uint8_t sht11_recv_byte(void)
{
   uint8_t result = 0;
   
   DATA_AS_IN();
   
   int i;
   for(i = 7; i >= 0; --i)
   {
      SCLK_HI();
         
      if (PORT_IN(SHT11_DATA_PORT) & (1 << SHT11_DATA_PIN))
	 result |= (1 << i);

      SCLK_LO();
   }

   return result;
}

/** @brief reset the comm. interface -- status register remains unchanged */  
static inline void sht11_reset_interface(void)
{
   DATA_AS_OUT();
   DATA_HI();
   SCLK_LO();
   
   int i;
   for(i = 0; i < 9; ++i)
   {
      SCLK_HI();
      SCLK_LO();
   }

   // must wait 11 ms after resetting the interface to send another command
   delay();
}

/** @brief send the transmission start sequence */
static inline void sht11_transmission_start(void)
{
   DATA_AS_OUT();

   DATA_HI();
   SCLK_LO();
   SCLK_HI();
   DATA_LO(); // falling edge on DATA
   SCLK_LO(); // low pulse on SCLK
   SCLK_HI();
   DATA_HI(); // rising edge on DATA
   SCLK_LO();
}

/** @brief returns TRUE if the SHT11 pulls data low to ack. */
static inline uint8_t sht11_recv_ack(void)
{
   uint8_t i;
   DATA_AS_IN();
   DATA_HI();
   
   SCLK_HI();
   i = ((PORT_IN(SHT11_DATA_PORT) & (1 << SHT11_DATA_PIN)) ? FALSE : TRUE);
   SCLK_LO();
   
   return i;
}

/** @brief pulls data low and pulses SCLK to ack reception of a byte from the SHT11 */
static inline void sht11_send_ack(void)
{
   DATA_AS_OUT();
   DATA_LO();

   SCLK_HI();
   SCLK_LO();
}

/** @brief sends a command to the SHT11 and returns the result, or 0xFFFF on failure */
static uint16_t sht11_perform_cmd(uint8_t command)
{
   
   uint16_t result = 0;
   sht11_transmission_start();
   sht11_send_byte(command);
   if (!sht11_recv_ack())
   {
      // return failure
      return (uint16_t)-1;
   }
   

   // process the response
   switch(command)
   {
   case SHT11_CMD_READ_TEMP:
   case SHT11_CMD_READ_HUMIDITY:
      // wait for measurement to complete     
      while(PORT_IN(SHT11_DATA_PORT) & (1 << SHT11_DATA_PIN));

      result |= ((uint16_t)sht11_recv_byte()) << 8;
      sht11_send_ack();
      result |= sht11_recv_byte();
      break;
   case SHT11_CMD_READ_SR:
      result = sht11_recv_byte();
      break;
   default:
      // write sr & soft reset do not return data
      break;
   }

   return result;
   
}

/** @brief disables power to the SHT11 */
static inline void sht11_power_off(void)
{
   UNSET_PORT(SHT11_POWER);
}

/** @brief enables power to the SHT11 */
static inline void sht11_power_on(void)
{
   SET_PORT(SHT11_POWER);
}

/** @brief initializes the SHT11 pins with power off */ 
void sht11_init(void)
{
   mode = DEV_MODE_OFF;
   
   // do not power the sensor yet.
   SET_DIR(SHT11_POWER);
   sht11_power_off();
   
   SET_DIR(SHT11_SCLK);
   SCLK_LO();
   DATA_AS_IN();
   DATA_HI();

}

/** @brief returns a uint16_t in buf containing the measured temperature */
uint8_t read_TEMP_sensor(void *buf, uint16_t count)
{
   if (mode == DEV_MODE_OFF)
      set_mode_TEMP_sensor(DEV_MODE_ON);
   
   if (count < 2)
      return NRK_ERROR;
   
   *((uint16_t*)buf) = sht11_perform_cmd(SHT11_CMD_READ_TEMP);
   return 2;
}


/** @brief turn the sensor on or off */
uint8_t set_mode_TEMP_sensor (uint8_t newMode)
{
   switch(newMode) {
   case DEV_MODE_ON:
   case DEV_MODE_IDLE:
      sht11_power_on();
      sht11_reset_interface();
      if (sht11_perform_cmd(SHT11_CMD_SOFT_RESET) == (uint16_t)-1)
           return NRK_ERROR;
      break;

   case DEV_MODE_OFF:
      sht11_power_off();
      break;

   default:
      return NRK_ERROR;
      break;
   }

   mode = newMode;
   return NRK_OK;
}



/*------------------light sensor functions-------------------------*/

void adc_init(){
	adc_off();
}
void adc_set_channel(uint8_t ch)
{
   uint8_t addr = 0;
   //set the addr where conversion result will reside
   ADC12CTL1 |= ((addr & 0x0F) << 11);
   //set the addr to read from the corresponding channel
   ADC12MCTL0 |= (ch & 0x0F);

}

uint16_t adc_get_conversion16(uint8_t ch)
{

   ADC12CTL0 &= ~ENC;

   adc_on();

   adc_set_channel(ch);
   
   //set source of sampling signal (SAMPCON) to be the output of
   //the sampling timer
   ADC12CTL1 |= SHP;

   //start the conversion
   ADC12CTL0 |= ENC | ADC12SC;

   //polling wait for conversion to complete
   while(ADC12CTL1 & ADC12BUSY);

   uint16_t sample = ADC12MEM0;
   return sample;
}



uint8_t dev_manager_tmote_sensors(uint8_t action,uint8_t opt,uint8_t *buffer,uint8_t size)
{
// key and value get passed as opt and size
uint8_t key=opt;
uint8_t value=size;

     switch(action)
     {
            case INIT: 
			// Set the pwr ctrl pin as output
			  sht11_init(); 
                          adc_init();
                       is_open=0;
		      return 1;
	     
	    case OPEN:  
		    if(is_open==1) return NRK_ERROR;
		    is_open=1; 
		    if(opt&READ_FLAG)
		    {
		   	// Turn on Sensor Node Power
			set_mode_TEMP_sensor(DEV_MODE_ON);
					    	return NRK_OK;
		    }
		    if(opt&WRITE_FLAG)
		    {
		    	return NRK_ERROR;
		    }
		    if(opt&APPEND_FLAG)
		    {
		    	return NRK_ERROR;
		    }
		    if(((opt)&(READ_FLAG|WRITE_FLAG|APPEND_FLAG))==0)
		    	return NRK_ERROR;
		    else return NRK_OK;
		
	    

             case READ:
		       
			if(size != 2) return 0;
			if(channel == TEMP || channel == HUMIDITY)
			{	//nrk_led_toggle(RED_LED);	
				 return read_TEMP_sensor(buffer, size); 
						        
			}

			
			if(channel == LIGHT){
				//nrk_led_toggle(GREEN_LED);
				uint16_t temp = adc_get_conversion16(5);
 				temp = CALC_LIGHT(temp);				
				//printf( "-%u \r\n",temp);				
				*buffer = temp;				
				return temp;
			}
				
			
             case CLOSE:
			// Turn off sensor power
                        set_mode_TEMP_sensor(DEV_MODE_OFF);
		    	is_open=0;
			adc_off(); 
                        return NRK_OK;
             
	     case GET_STATUS:
	     		// use "key" here 
			if(key==SENSOR_SELECT) return channel;
	     		return NRK_ERROR;
			
             case SET_STATUS:
	     		// use "key" and "value" here
					
			if(key==SENSOR_SELECT) 
			{
			if(value==TEMP || value == HUMIDITY||value == LIGHT) 
			  {	//nrk_led_set(RED_LED);  	
				channel=value;
				return NRK_OK;

			  }
			
			}
			return NRK_ERROR;
	     default:
		nrk_kernel_error_add(NRK_DEVICE_DRIVER,0);
		 return 0;
	}
}

void delay()
{
  nrk_spin_wait_us(SENSOR_SETUP_DELAY);
}


void enable_user_button()
{
   nrk_led_toggle(RED_LED);
       
   P2DIR &= ~(1 << 7);
   P2SEL &= ~(1 << 7);
   P2IES &= ~(1 << 7);
   P2IE |= (1 << 7);
   BUTTON_STATUS = B_ON;	
}

interrupt (PORT2_VECTOR) user_button_interrupt()
{
		
   if (P2IFG & (1 << 7))
   {
      P2IFG &= ~(1 << 7);
      if(BUTTON_STATUS == B_ON){
		BUTTON_STATUS = B_OFF;
      }
      else{
		BUTTON_STATUS = B_ON;
      }		
      
   }
	
}
